import os
from dotenv import load_dotenv
load_dotenv()
# Import classes from sibling modules (package-style)
from .decompose import DecomposeQuestion, QuestionFilter
from .chunk_extractor import ChunkExtaction
from .reranker import ReRanker
from .final_llm import final_query

base_url = "http://localhost:9018/v1"
model = "ibnzterrell/Meta-Llama-3.3-70B-Instruct-AWQ-INT4"

def main():
    user_question = """
    explain Plancks Quantum Hypothesis.
    """


    print("\n--- Processing your question, please wait... ---")

    # Decomposition
    decomposer = DecomposeQuestion(model=model, base_url=base_url)
    print("\n--- Generating Sub-questions ---")
    sub_questions = decomposer.decompose_function(user_question)
    # print("Generated Sub-questions:\n", sub_questions)

    # Filter sub-questions
    question_filter = QuestionFilter(model=model, base_url=base_url)
    print("\n--- Filtering Sub-questions ---")
    filtered_questions = question_filter.filter_subquestions_function(user_question, sub_questions)
    print("Filtered Top 5 Sub-questions:\n", filtered_questions)

    # Sub-questions chunk extractors
    ChunkExtactorInstance = ChunkExtaction()
    print("\n--- Embedding Sub-questions and Extracting Chunks ---")
    filtered_questions_chunks = ChunkExtactorInstance.chunker_function(filtered_questions)
    print("\n--- Chunks Found ---")

    # Chunk re-ranking
    Ranker = ReRanker()
    print("\n--- Re Ranking ---")
    final_ranked_list = Ranker.reranker_function(filtered_questions_chunks)
    print("\n--- Re Ranking Done ---")

    # Feed to LLM
    llm_caller = final_query(model=model, base_url=base_url)
    print("\n--- Querying LLM --- \n")
    FinalAnswers = llm_caller.caller_function(final_ranked_list)
    print(FinalAnswers)

    # for i, answer in enumerate(FinalAnswers):
    #     print("\n -------------- \nans " , i , " : " , answer, "\n -------------- \n" )
    print("ALL DONE, WOHOO")

if __name__ == "__main__":
    main()



from langchain_community.vectorstores import FAISS
from sentence_transformers import SentenceTransformer
from langchain_core.embeddings import Embeddings
import json


class SentenceTransformerEmbeddings(Embeddings):
    def __init__(self, model_name):
        self.model = SentenceTransformer(model_name, device="cpu")
        
    def embed_documents(self, texts):
        return [self.embed_query(text) for text in texts]
    
    def embed_query(self, text):
        return self.model.encode(text).tolist()
    

class ChunkExtaction:

    def __init__(self):
        embedding_model = SentenceTransformerEmbeddings("all-mpnet-base-v2")
        self.library = FAISS.load_local("faiss_index", embedding_model, allow_dangerous_deserialization=True)
    
    """
    def chunker_function(self, filtered_questions):  
        query_answers = []
        #iterate through 5 q
        print("Number of input questions:", len(filtered_questions))
        for i, q in enumerate(filtered_questions[:5]):
            # 5 chunks for each q
            document_chunks  = self.library.similarity_search(q, k=5)
            string_chunks = [doc.page_content for doc in document_chunks]   
            #dict def per q
            print("\n ------- len of string_chunks:" , len(string_chunks))
            result_item = {
                "index": i,
                "question": q,
                "search_results": string_chunks
            }
            query_answers.append(result_item)
        print("\n --------- len of query_answers:", len(query_answers))
        return query_answers

    """
    def chunker_function(self, filtered_questions_json):

        input_list = json.loads(filtered_questions_json)
        query_answers = []
        print("Number of input questions:", len(input_list))
        for i, item in enumerate(input_list):
            q_number = item.get("q_number")
            q = item.get("q")
            # 5 chunks for each q
            document_chunks  = self.library.similarity_search(q, k=5)
            string_chunks = [doc.page_content for doc in document_chunks]
            #print(f"\n ------- len of string_chunks for q_number {q_number}:", len(string_chunks))
            result_item = {
                "index": i,
                "q_number": q_number,
                "question": q,
                "search_results": string_chunks
            }
            query_answers.append(result_item)
        #print("\n --------- len of query_answers:", len(query_answers))
        return query_answers



from langchain_community.vectorstores import FAISS
from sentence_transformers import SentenceTransformer
from langchain_core.embeddings import Embeddings
from openai import OpenAI
from google.genai import types, errors
import os
from dotenv import load_dotenv
load_dotenv()

class DecomposeQuestion:

    def __init__(self, base_url, model):
        self.base_url = base_url
        self.model = model
        self.client = OpenAI(base_url=self.base_url)

    def decompose_function(self, orignal_query):
        # Prompt
        prompt_template = """
        Your task is to analyse the providing question then raise atomic sub-questions for the knowledge that can help you answer the question better.
        Think in different ways and raise as many diverse questions as possible. The user will provide the base question. 
        Do not answer the question, just raise sub-questions. Do not decompose if the question is simple enough
        """

        # API call
        try:
            response = self.client.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": prompt_template},
                    {"role": "user", "content": orignal_query},
                ]
            )
            #print( "Response : \n ",response.choices[0].message.content, "Response over \n")
            return response.choices[0].message.content
        except errors.APIError as e:
            print(e.code)
            print(e.message)


class QuestionFilter:

    def __init__(self, base_url, model):
        self.base_url = base_url
        self.model = model
        self.client = OpenAI(base_url=self.base_url)

    def filter_subquestions_function(self, orignal_query, subquestions):
        #  Prompt
        prompt_template = """
        You will receive an original question and a list of sub-questions.
        Your tasks:
        1. Compare each sub-question to the original question for relevance.
        2. Select only the sub-questions that are directly useful to answer the original question.
        3. Return 5 sub-questions as a numbered list. Do not repeat, elaborate, or answer.
        IMPORTANT - If the orignal question is simple enough, select even lesser sub-questions.

        Output Format:
        Example output:
        [
        {"q_number": 1, "q": "First selected sub-question"},
        {"q_number": 2, "q": "Second selected sub-question"}
        ]
        """

        # API call
        try:
            response = self.client.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": prompt_template},
                    {"role": "user", "content": f"Original question: {orignal_query}\nSub-questions:\n{subquestions}"},
                ]
            )
            #print("Response \n",response.choices[0].message.content,  "Response over \n")
            return response.choices[0].message.content
        except errors.APIError as e:
            print(e.code)
            print(e.message)